#include <avr/io.h>

// read : 08 00 00 00 01 00
// write: 0a 00 00 00 01 00

// PORTA	0x1B
// PINA		0x19
#define DATA_OUT	PORTA
#define DATA_IN		PINA
#define DATA_DIR	DDRA

// PORTC	0x15
// PINC		0x13
#define CTRL_OUT	PORTC
#define CTRL_IN		PINC
#define CTRL_DIR	DDRC

#define BYTE	unsigned char
#define WORD	unsigned short

//PORT A -- ACSI DATA
// -- pozor, reversed (PA7 == D0, ..., PA0 == D7)

// PORT C
#define	aRESET	(1<<6)
#define aACK	(1<<5)
#define aA1		(1<<4)
#define aRW		(1<<3)
#define aDRQ	(1<<2)
#define aIRQ	(1<<1)
#define aCS	    1

#define bCS		0
#define bACK	5
//---------------------
#define	IDLE		0x7f

#define CMD_FIRST	0x66
#define CMD_REST	0x76
#define STATUS		0x7e

#define	DMA_WRITE	0x57
#define DMA_READ	0x5f
//---------------------

#define ASM_CONSTANTS asm(	".equ DATA_OUT,  0x1b"	"\n\t"	\
".equ DATA_IN,   0x19"	"\n\t"	\
".equ CTRL_OUT,  0x15"	"\n\t"	\
".equ CTRL_IN,   0x13"	"\n\t"	\
".equ IDLE,      0x7f"	"\n\t"	\
".equ CMD_FIRST, 0x66"	"\n\t"	\
".equ CMD_REST,  0x76"	"\n\t"	\
".equ STATUS,    0x7e"	"\n\t"	\
".equ DMA_WRITE, 0x57"	"\n\t"	\
".equ DMA_READ,  0x5f"	"\n\t"	\ 
".equ bCS,		 0"		"\n\t"	\
".equ bACK,		 5"		"\n\t" );

#define SET(rgstr, flags)        ((rgstr) = (rgstr) | (flags))
#define CLEAR(rgstr, flags)      ((rgstr) &= ~(flags))
#define TOGGLE(rgstr, flags)     ((rgstr) ^= (flags))

#define LED_ACT	(1<<4)

#define LED_ON()	CLEAR(PORTB, LED_ACT)
#define LED_OFF()	SET(PORTB, LED_ACT)

#define TIMEOUT		60000
//---------------------

BYTE read_value;

void pio_w1(BYTE val);
BYTE pio_w(BYTE val);
BYTE pio_r();

BYTE dma_w(BYTE val);
BYTE dma_r();

void writeSector();
void readSector();

#define dataAsInputs()	(DATA_DIR = 0x00)
#define dataAsOutputs()	(DATA_DIR = 0xff)

#define reverse(IN)		(tab[0xff & IN])

void uart_init();
void uart_puts(char *str);
void uart_putNumber(int val);
void uart_putHexa(BYTE val);
void uart_putchar(char data);
BYTE uart_getchar();
BYTE uart_gotchar();

void dump(BYTE *data);
void backwardArray(BYTE *arr);
BYTE backward(BYTE val);
//---------------------
BYTE tab[256] = {0,128,64,192,32,160,96,224,16,144,80,208,48,176,112,240,8,136,72,200,40,168,104,
232,24,152,88,216,56,184,120,248,4,132,68,196,36,164,100,228,20,148,84,212,52,180,
116,244,12,140,76,204,44,172,108,236,28,156,92,220,60,188,124,252,2,130,66,194,34,
162,98,226,18,146,82,210,50,178,114,242,10,138,74,202,42,170,106,234,26,154,90,
218,58,186,122,250,6,134,70,198,38,166,102,230,22,150,86,214,54,182,118,246,14,
142,78,206,46,174,110,238,30,158,94,222,62,190,126,254,1,129,65,193,33,161,97,
225,17,145,81,209,49,177,113,241,9,137,73,201,41,169,105,233,25,153,89,217,57,185,
121,249,5,133,69,197,37,165,101,229,21,149,85,213,53,181,117,245,13,141,77,205,
45,173,109,237,29,157,93,221,61,189,125,253,3,131,67,195,35,163,99,227,19,147,83,
211,51,179,115,243,11,139,75,203,43,171,107,235,27,155,91,219,59,187,123,251,7,
135,71,199,39,167,103,231,23,151,87,215,55,183,119,247,15,143,79,207,47,175,111,
239,31,159,95,223,63,191,127,255};

//---------------------
int main(void)
{
	// init pins to known state
	CTRL_DIR = aRESET | aACK | aA1 | aRW | aCS;			// set these as outputs
	CTRL_OUT = IDLE;										// and set CTRL signals to IDLE state
	
	TWCR = 0; //disable twi
		
	// activity LED
	DDRB = LED_ACT;	
	
	dataAsInputs();
	// end of init
	
	uart_init();
	uart_puts("DMA sim up!\n\rPress w or r.\n\r");
	
	LED_OFF();
	
	while(1) {
		if(uart_gotchar()) {
			BYTE ch = uart_getchar();
			
			if(ch == 'w' || ch == 'W') {
				LED_ON();
				writeSector();
				LED_OFF();
			}
			
			if(ch == 'r' || ch == 'R') {
				LED_ON();
				readSector();
				LED_OFF();
			}
			
			if(ch == 'i' || ch == '?') {
				LED_ON();
				uart_puts("Press w or r.\n\r");
				LED_OFF();
			}					
		}
	}
}
//----------------------------------------------------------------------------------
void readSector()
{	// read: 08 00 00 00 01 00
	BYTE cmds[5] = {0, 0, 0, 1, 0};
	int i;

	BYTE data[512];
	
	uart_puts("read: ");
	//-------------------
	// send commands in PIO mode
	pio_w1(0x08);						// send 1st CMD byte

	for(i=0; i<5; i++) {				// try to send remaining 5 CMD bytes
		if(pio_w(cmds[i]) == 0) {		// if sending failed, quit
			dataAsInputs();
			
			uart_puts("fail cmd ");
			uart_putNumber(i);
			uart_puts("\n\r");
			
			return;
		}
	}	
	//-------------------
	// get data in DMA mode	
	dataAsInputs();
	
	for(i=0; i<512; i++) {
		if(dma_r() == 0) {				// if reading failed, quit
			
			uart_puts("fail dma ");
			uart_putNumber(i);
			uart_puts("\n\r");

			return;
		}				
		
		data[i] = read_value;
	}
	//-------------------
	// read status in PIO mode
	if(pio_r() == 0 || read_value != 0)	{	// if status byte read failed or bad status returned
		uart_puts("fail status\n\r");
	}

	for(i=0; i<512; i++) {
		if(data[i] != ((BYTE) i)) {
			uart_puts(" fail check ");
			uart_putHexa(i);
			uart_puts(" - ");
			uart_putHexa(data[i]);
			uart_puts("\n\r");
			
			dump(data);
			
			return;
		}
	}					

	uart_puts(" OK!\n\r");
}
//----------------------------------------------------------------------------------
void writeSector() 
{	//write: 0a 00 00 00 01 00
	BYTE cmds[5] = {0, 0, 0, 1, 0};
	int i;
		
	uart_puts("write: ");

	//-------------------
	// send commands in PIO mode
	pio_w1(0x0a);						// send 1st CMD byte
	
	for(i=0; i<5; i++) {				// try to send remaining 5 CMD bytes
		if(pio_w(cmds[i]) == 0) {		// if sending failed, quit
			dataAsInputs();
			
			uart_puts("fail cmd ");
			uart_putNumber(i);
			uart_puts("\n\r");

			return;
		}
	}	
	//-------------------
	// send data in DMA mode	
	for(i=0; i<512; i++) {
		if(dma_w(i) == 0) {		// if sending failed, quit
			dataAsInputs();
			
			uart_puts("fail dma ");
			uart_putNumber(i);
			uart_puts("\n\r");
				
			return;
		}				
	}		

	//-------------------
	// read status in PIO mode
	if(pio_r() == 0 || read_value != 0)	{	// if status byte read failed or bad status returned
		uart_puts("fail status\n\r");
	}		
	
	uart_puts(" OK!\n\r");
}
//----------------------------------------------------------------------------------
void dump(BYTE *data)
{
	int c=0;
	BYTE r,s;

	uart_puts("\n\r\n\rdump:\n\r");
	
	for(r=0; r<16; r++) {
		for(s=0; s<32; s++) {
			uart_putHexa(data[c]);
			uart_putchar(' ');
			c++;
		}
		
		uart_puts("\n\r");
	}
}	
//----------------------------------------------------------------------------------
void backwardArray(BYTE *arr)
{
	int i;
	for(i=0; i<512; i++) {
		arr[i] = backward(arr[i]);
	}	
}
//----------------------------------------------------------------------------------
BYTE backward(BYTE val)
{
	int i;
	for(i=0; i<256; i++) {
		if(val == tab[i]) {
			return i;
		}
	}	
	
	return 0xff;
}
//----------------------------------------------------------------------------------
void pio_w1(BYTE val) 
{
	dataAsOutputs();
	
	val = reverse(val);
	DATA_OUT = val;
	
	ASM_CONSTANTS;
	
asm("LDI R24, IDLE"			"\n\t"		// IDLE
	"LDI R25, CMD_FIRST"	"\n\t"		// CMD_FIRST

	"OUT CTRL_OUT,R24"		"\n\t"		// set IDLE
	"OUT CTRL_OUT,R25"		"\n\t"		// set CMD_FIRST

	"NOP"					"\n\t"		
	"NOP"					"\n\t"		
	"NOP"					"\n\t"		
	
	"SBI CTRL_OUT,bCS"		"\n\t"		// first return CS to H
	"NOP"					"\n\t"

	"OUT CTRL_OUT,R24"		"\n\t" );	
}

BYTE pio_w(BYTE val)
{
	BYTE in;
	WORD timeout = TIMEOUT;			// this could be couple of milliseconds
	
	while(1) {
		in = CTRL_IN;				// get CTRL
		if((in & aIRQ) == 0) {		// is IRQ down? good
			break;
		}
		
		if(timeout == 0) {			// time out? bad
			return 0;
		}			
		
		timeout--;
	}
	
	val = reverse(val);
	DATA_OUT = val;
	
	ASM_CONSTANTS;
	
asm("LDI R24, IDLE"			"\n\t"		// IDLE
	"LDI R25, CMD_REST"		"\n\t"		// CMD_REST

	"OUT CTRL_OUT,R24"		"\n\t"		// set IDLE
	"OUT CTRL_OUT,R25"		"\n\t"		// set CMD_REST

	"NOP"					"\n\t"
	"NOP"					"\n\t"
	"NOP"					"\n\t"

	"SBI CTRL_OUT,bCS"		"\n\t"		// first return CS to H
	"NOP"					"\n\t"

	"OUT CTRL_OUT,R24"		"\n\t");
	
	return 1;
}

BYTE pio_r()
{
	dataAsInputs();
	
	BYTE in;
	WORD timeout = TIMEOUT;			// this could be couple of milliseconds
	
	while(1) {
		in = CTRL_IN;				// get CTRL
		if((in & aIRQ) == 0) {		// is IRQ down? good
			break;
		}
	
		if(timeout == 0) {			// time out? bad
			return 0;
		}

		timeout--;
	}
	
	ASM_CONSTANTS;

asm("LDI R24, IDLE"			"\n\t"		// IDLE
	"LDI R25, STATUS"		"\n\t"		// STATUS		

	"OUT CTRL_OUT,R24"		"\n\t"		// set IDLE
	"OUT CTRL_OUT,R25"		"\n\t"		// set STATUS

	"NOP"					"\n\t"
	"NOP"					"\n\t"
	"IN R25, DATA_IN"		"\n\t"		// R25 = DATA_IN

	"SBI CTRL_OUT,bCS"		"\n\t"		// first return CS to H
	"NOP"					"\n\t"

	"OUT CTRL_OUT, R24"		"\n\t"		// set IDLE
	"STS read_value, R25"	"\n\t" );	// acsi_status = R25

	read_value = reverse(read_value);

	return 1;
}

BYTE dma_w(BYTE val)
{
	BYTE in;
	WORD timeout = TIMEOUT;			// this could be couple of milliseconds
	
	while(1) {
		in = CTRL_IN;				// get CTRL
		if((in & aDRQ) == 0) {		// is IRQ down? good
			break;
		}
	
		if(timeout == 0) {			// time out? bad
			return 0;
		}

		timeout--;
	}
	
	val = reverse(val);
	DATA_OUT = val;
	
	ASM_CONSTANTS;
	
asm("LDI R24,IDLE"		"\n\t"		// IDLE
	"LDI R25,DMA_WRITE"	"\n\t"		// DMA_WRITE

	"OUT CTRL_OUT,R24"	"\n\t"		// set IDLE
	"OUT CTRL_OUT,R25"	"\n\t"		// set DMA_WRITE

	"NOP"				"\n\t"
	"NOP"				"\n\t"
	"NOP"				"\n\t"

	"SBI CTRL_OUT,bACK"		"\n\t"		// first return ACK to H
	"NOP"					"\n\t"

	"OUT CTRL_OUT,R24"	"\n\t");	// set IDLE
	
	return 1;
}

BYTE dma_r()
{
	BYTE in;
	WORD timeout = TIMEOUT;			// this could be couple of milliseconds
	
	while(1) {
		in = CTRL_IN;				// get CTRL
		if((in & aDRQ) == 0) {		// is IRQ down? good
			break;
		}
	
		if(timeout == 0) {			// time out? bad
			return 0;
		}

		timeout--;
	}
	
	ASM_CONSTANTS;
	
asm("LDI R24,IDLE"			"\n\t"		// IDLE
	"LDI R25,DMA_READ"		"\n\t"		// DMA_READ

	"OUT CTRL_OUT,R24"		"\n\t"		// set IDLE
	"OUT CTRL_OUT,R25"		"\n\t"		// set DMA_READ

	"NOP"					"\n\t"
	"NOP"					"\n\t"
	"IN R25, DATA_IN"		"\n\t"		// R25 = DATA_IN

	"SBI CTRL_OUT,bACK"		"\n\t"		// first return ACK to H
	"NOP"					"\n\t"

	"OUT CTRL_OUT,R24"		"\n\t"
	"STS read_value, R25"	"\n\t");	// set IDLE
	
	read_value = reverse(read_value);
	
	return 1;
}

void uart_init()
{
	// bauds: 19'200
	UBRRL = 51;
	UBRRH = 0;

	// async mode, No Parity,  1 StopBit,  char size 8
	UCSRC=(1<<URSEL)|(3<<UCSZ0);

	//Enable The receiver and transmitter
	UCSRB=(1<<RXEN)|(1<<TXEN);
}

void uart_puts(char *str) 
{
	while(1) {
		BYTE val = *str;
		
		if(val == 0) {
			return;
		}
		
		uart_putchar(val);
		str++;
	}	
}

void uart_putNumber(int val) 
{
	BYTE s,d,j, tmp;
	s	= val / 100;
	tmp = val % 100;
	d	= tmp /  10;
	j	= tmp %  10;
	
	s += 48;
	d += 48;
	j += 48;
	
	uart_putchar(s);
	uart_putchar(d);
	uart_putchar(j);
}

void uart_putHexa(BYTE val)
{
	BYTE hi, lo;
	char hex[16] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};
	
	hi = val >> 4;
	lo = val & 0x0f;
	
	char h,l;
	h = hex[hi];
	l = hex[lo];
	
	uart_putchar(h);
	uart_putchar(l);
}

void uart_putchar(char data)
{
	// Make sure that the UART buffer is empty
	while (!(UCSRA & (1 << UDRE)));
	UDR = data;
}

BYTE uart_getchar()
{
	while (!(UCSRA & (1 << RXC)));
	
	return UDR;
}

BYTE uart_gotchar()
{
	if(UCSRA & (1 << RXC)) {
		return 1;
	}
	
	return 0;
}